import numpy as np


def generate_drift(df, drift_list=[1.1, 10.0, 0.1, 1.1]):
    drift_array = np.array(drift_list)
    quotient, remain = np.divmod(df.shape[0], len(drift_array))

    quotient_array = np.repeat(drift_list, quotient)
    remain_array = np.repeat(drift_list[-1], remain)

    drift = np.append(quotient_array, remain_array)
    return drift



import matplotlib.pyplot as plt


def plot_multiple_instances_results(
    alg_dict,
    y_label,
    marker=".",
    alpha=0.7,
    default=None,
    log_x=False,
    log_y=False,
    title="",
):

    fig, ax = plt.subplots(figsize=(14, 8))

    if default is not None:
        ax.axhline(y=default, color="black", linestyle="-.", label="default")

    for i, (algo_k, algo_v) in enumerate(alg_dict.items()):
        ax.plot(
            algo_v.keys(),
            algo_v.values(),
            alpha=alpha,
            marker=marker,
            linestyle="dashed",
            label=algo_k,
        )

    ax.set_title = title
    if log_x:
        ax.set_xscale("log")
    if log_y:
        ax.set_yscale("log")
    ax.set_ylim(0)
    ax.grid(True)
    ax.set_xlabel("Number of train observations")
    ax.set_ylabel(y_label)
    ax.legend()
    plt.show()



import hashlib
from pathlib import Path

def sha256sum(filename):
    hash = hashlib.sha256()
    buf = bytearray(1024 * 1024) # Reusable 1MB buffer
    view = memoryview(buf)
    with open(filename, "rb") as fd:
        while True:
            size = fd.readinto(buf)
            if size == 0:
                break  # EOF
            hash.update(view[:size])
        return hash.hexdigest()

def get_data_home(data_home=None) -> str:
    """Return the path of the pyspotstream data directory.
    This folder is used by some large dataset loaders to avoid downloading the
    data several times.
    By default the data directory is set to a folder named 'pyspotstream_data' in the
    user home folder.
    Alternatively, it can be set by the 'PYSPOTSTREAM_DATA' environment
    variable or programmatically by giving an explicit folder path. The '~'
    symbol is expanded to the user home folder.
    If the folder does not already exist, it is automatically created.
    
    Parameters
    ----------
    data_home : str, default=None
        The path to pyspotstream data directory. If `None`, the default path
        is `~/pyspotstream_data`.
        
    Returns
    -------
    data_home: str
        The path to the pyspotstream data directory.
    """
    if data_home is None:
        data_home = environ.get("PYSPOTSTREAM_DATA", 
                                Path.home() / "pyspotstream_data")
    # Ensure data_home is a Path() object pointing to an absolute path
    data_home = Path(data_home).absolute()
    # Create data directory if it does not exists.
    data_home.mkdir(parents=True, exist_ok=True)
    
    return data_home




import time
import datetime as dt
import matplotlib.pyplot as plt

from sklearn import tree
from sklearn.model_selection import cross_val_score
from river import evaluate


class plotPerformance:
    """Basic plot with three subplots: error, time, and memory"""

    def __init__(self, start=0, end=None, figsize=(10, 5)):
        self.figsize = figsize
        self.nrows = 3
        self.dpi = 300
        self.start = start
        self.end = end

    def plot_performance(self, dataset, metric, models):
        metric_name = metric.__class__.__name__

        # To make the generated data reusable
        dataset = list(dataset)[self.start : self.end]
        fig, ax = plt.subplots(figsize=self.figsize, nrows=self.nrows, dpi=self.dpi)
        for model_name, model in models.items():
            step = []
            error = []
            r_time = []
            memory = []

            for checkpoint in evaluate.iter_progressive_val_score(
                dataset, model, metric, measure_time=True, measure_memory=True, step=100
            ):
                step.append(checkpoint["Step"])
                error.append(checkpoint[metric_name].get())

                # Convert timedelta object into seconds
                r_time.append(checkpoint["Time"].total_seconds())
                # Make sure the memory measurements are in MB
                raw_memory = checkpoint["Memory"]
                memory.append(raw_memory * 2**-20)

            ax[0].plot(step, error, label=model_name)
            ax[1].plot(step, r_time, label=model_name)
            ax[2].plot(step, memory, label=model_name)

        ax[0].set_ylabel(metric_name)
        ax[1].set_ylabel("Time (seconds)")
        ax[2].set_ylabel("Memory (MB)")
        ax[2].set_xlabel("Instances")

        ax[0].grid(True)
        ax[1].grid(True)
        ax[2].grid(True)

        ax[0].legend(
            loc="upper center",
            bbox_to_anchor=(0.5, 1.25),
            ncol=3,
            fancybox=True,
            shadow=True,
        )
        plt.tight_layout()
        plt.close()

        return fig


def model_memory_eval(obj, unit="byte", protocol=3):
    """
    Model size estimation with pickle
    Parameters
    ----------
    obj: python object

    unit : str, default='byte'
        'kB'
        'MB'

    protocol : int, default=3
    """
    import pickle
    import sys

    size = sys.getsizeof(pickle.dumps(obj, protocol=protocol))
    # print('pickle size:',size , 'byte')
    if unit == "kB":
        return size / 1024

    if unit == "MB":
        return size / (2**20)

    return size


def batch_tree_performance(x, y):
    """Estimate performance (error, time, and memory usage) for a Decision Tree Classifier from the sklearn package"""

    # To evaluate time and memory usage, we train a Decision Tree Classifier (DTC) on the entire dataset.
    time_start = time.perf_counter()
    dtc = tree.DecisionTreeClassifier()
    dtc = dtc.fit(x, y)
    time_now = time.perf_counter()
    time_delta = dt.timedelta(seconds=time_now - time_start)
    memory_usage = model_memory_eval(dtc)

    # Evaluate a model with 5-fold cross-validation
    scores = cross_val_score(tree.DecisionTreeClassifier(), x, y, cv=5)
    mean_score = scores.mean()

    # Return time, memory usage and model error
    return time_delta.microseconds / 1000, memory_usage / 1024, mean_score



import gc
import time
import tracemalloc

import numpy as np
from sklearn import metrics as sk_metrics
from sklearn.model_selection import train_test_split
from river import stream as river_stream
from river import preprocessing as river_preprocessing


def get_acronym(x):
    """
    Maps long name to short acronym, e.g., `"mean_absolute_error"` to `"MAE"`.

    Args:
        x (string): long name

    Returns:
        string: acronym
    """
    if x == "mean_absolute_error":
        return "MAE"
    else:
        return "Acronym unknown."


def eval_batch_machine_learning(
    X,
    Y,
    model,
    random_state=0,
    test_size=0.90,
    shuffle=False,
    metric=sk_metrics.mean_absolute_error,
    verbosity=True,
):
    """
    Standard batch machine learning experiment. The `train_test_split` method from
    `sklearn.model_selection` is applied to the input data set (`X` and `y`).

    Args:
        X (DataFrame): Pandas DataFrame that contains the input values.
        y (Series): Pandas Series that contains the output values.
        model: Batch Learner (model), e.g., `DecisionTreeRegressor` from `sklearn`.
        random_state (int): Controls the shuffling applied to the data before applying
        the split. Only needed if `shuffle=True`.
        Pass an int for reproducible output across multiple function calls.
        See :term:`Glossary <random_state>`.
        test_size (float or int):
            If float, should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the test split. If int, represents the
            absolute number of test samples. If None, the value is set to the
            complement of the train size. If ``train_size`` is also None, it will
            be set to 0.25.
        shuffle (bool):
            Whether or not to shuffle the data before splitting. Note: In contrast to
            `sklearn.model_selection`'s default setting (which is `True`), the default
            is `False`.
        metric: Defaults to sk_metrics.mean_absolute_error.
        verbosity (bool): If `True`, metrics are printed.

    Returns:
        dict: model_times
        dict: model_scores
        dict: model_mem
        model: model
    """

    train_X, test_X, train_y, test_y = train_test_split(
        X, Y, test_size=test_size, random_state=random_state, shuffle=shuffle
    )

    gc.collect()
    tracemalloc.start()
    tracemalloc.reset_peak()
    tic = time.time()
    model.fit(train_X, train_y)

    model_batch_time = time.time() - tic
    model_batch_mem = tracemalloc.get_traced_memory()[1] / 1024

    pred_y = model.predict(test_X)
    model_batch_mae = metric(test_y, pred_y)
    metric_name = str(metric.__name__)

    if verbosity:
        print(f"{get_acronym(str(metric.__name__))} of {str(model.__class__.__name__)}: ", model_batch_mae)
        print(f"Time of {str(model.__class__.__name__)}: ", model_batch_time)
        print(f"Memory of {str(model.__class__.__name__)}: ", model_batch_mem)

    return model_batch_time, model_batch_mae, model_batch_mem, model


def baseline_batch_cross_val_experiment(
    X,
    Y,
    model,
    random_state=0,
    test_size=0.2,
    shuffle=False,
    metric=sk_metrics.mean_absolute_error,
    verbosity=True,
):
    """Not Working yet!!!.

    Args:
        X (DataFrame): Pandas DataFrame that contains the input values.
        y (Series): Pandas Series that contains the output values.
        model: Batch Learner (model), e.g., `DecisionTreeRegressor` from `sklearn`.
        random_state (int, optional): _description_. Defaults to 0.
        test_size (float, optional): _description_. Defaults to 0.2.
        shuffle (bool, optional): _description_. Defaults to False.
        metric (_type_, optional): _description_. Defaults to sk_metrics.mean_absolute_error.
        verbosity (bool, optional): _description_. Defaults to True.

    Returns:
        dict: model_times
        dict: model_scores
        dict: model_mem
        model: model
    """
    from sklearn.model_selection import cross_val_score
    from sklearn.model_selection import train_test_split
    from sklearn.metrics import make_scorer

    scoring = make_scorer(metric)
    train_X, test_X, train_y, test_y = train_test_split(
        X, Y, test_size=test_size, random_state=random_state, shuffle=shuffle
    )

    gc.collect()
    tracemalloc.start()
    tracemalloc.reset_peak()
    tic = time.time()
    # model.fit(train_X, train_y)
    cross_val_scores = cross_val_score(model, train_X, train_y, cv=5, scoring=scoring)
    model_batch_time = time.time() - tic
    model_batch_mem = tracemalloc.get_traced_memory()[1] / 1024
    #
    mean_cross_val_scores = cross_val_scores.mean()
    # or !
    # cross_validate
    #
    pred_y = model.predict(test_X)
    model_batch_mae = metric(test_y, pred_y)

    if verbosity:
        print(f"Cross val. score of {model}: ", mean_cross_val_scores)
        print(f"MAE of {model}: ", model_batch_mae)
        print(f"Time of {model}: ", model_batch_time)
        print(f"Memory of {model}: ", model_batch_mem)

    return model_batch_time, mean_cross_val_scores, model_batch_mem, model


def eval_mini_batch_machine_learning(
    X,
    y,
    model,
    x_part="linspace",
    m_sklearn=True,
    metric=sk_metrics.mean_absolute_error,
    eval_on_full_data=False,
    fit_on_available_data=False,
    fit_on_fixed=False,
    fixed_train_size=0,
    n_fit=10,
    n_splits=100,
    # return_as_df = False
    verbose=False,
):
    """This method executes a mini-batch experiment.

    Args:
        X (DataFrame): Pandas DataFrame that contains the input values.
        y (Series): Pandas Series that contains the output values.
        model: Batch Learner (model), e.g., `DecisionTreeRegressor` from `sklearn`.
        x_part (str, optional):
            Partition of the input space. Defaults to 'linspace'.
        m_sklearn (bool, optional):
            Indicator for `sklearn`models. If `False`, then `river` models are used.
        metric:
            Evaluation metric. Defaults to sk_metrics.mean_absolute_error.
        eval_on_full_data (bool, optional):
            Evaluation uses the full data set. Defaults to False.
        fit_on_available_data (bool, optional):
            Fit the model on the whole set of observed data.
        fit_on_fixed (bool, optional):
            Fit the model on a fixed data set. Defaults to False.
        fixed_train_size (int, optional):
            Size of the fixed data set. Defaults to 0.
        n_fit (int, optional):
            Fit (train) on the last n_fit partitions only.
            This is a moving window of size (n_fit x partition size, which is computed via `x_part`). Defaults to 10.
        n_splits (int, optional):
            Number of splits (mini-batches) used for `x_part`. Defaults to 100.
        verbose (bool, optional):
            verbosity level. Defaults to False.

    Returns:
        dict: model_times
        dict: model_scores
        dict: model_mem
        dict: model_dict
    """
    model_times = {}
    model_scores = {}
    model_mem = {}

    if x_part == "logspace":
        x_seq = np.logspace(
            0.5, 1, n_splits, base=X.shape[0], dtype=int, endpoint=eval_on_full_data
        )[1:]
    elif x_part == "linspace":
        x_seq = np.linspace(
            1, X.shape[0], n_splits, dtype=int, endpoint=eval_on_full_data
        )[1:]
    else:
        raise Exception(f"{x_part} is an invalid Argument for x_part")

    if verbose:
        print("Length of Trainings Data", X.shape[0])
    if verbose:
        print(f"x_seq: {x_seq}")

    model_dict = {}
    tracemalloc.start()

    if m_sklearn:
        # Sklearn Model: Start of Train and Eval Loop
        if verbose:
            print("Starting Loop for Sklearn Model")
        for i, break_point in enumerate(x_seq):

            model_dict[break_point] = model
            if verbose:
                print(f"{i}. Breaking Point: {break_point}")

            gc.collect()
            tracemalloc.reset_peak()
            tic = time.time()

            # -- TRAINING --
            if fit_on_fixed:
                # train on a fixed data set of size 0:fixed_train_size
                if verbose:
                    print(f"\tFit on 0:{fixed_train_size}.")
                model_dict[break_point].fit(
                    X.iloc[:fixed_train_size], y.iloc[:fixed_train_size]
                )

            elif fit_on_available_data or i < n_fit:
                # train on the full set of seen data, that is available until train_size, i.e., from 0 to train_size:
                if verbose:
                    print(f"\tFit on 0:{break_point}.")
                model_dict[break_point].fit(X.iloc[:break_point], y.iloc[:break_point])

            else:
                # train on the last n_fit partitions only. This is a moving window of size (n_fit x partition size).
                if verbose:
                    print(f"\tFit on {x_seq[i - n_fit]}:{break_point}.")
                model_dict[break_point].fit(
                    X.iloc[x_seq[i - n_fit] : break_point],
                    y.iloc[x_seq[i - n_fit] : break_point],
                )

            # -- EVALUATION --
            if eval_on_full_data:
                # predict and evaluate on the full data set:
                if verbose:
                    print("\tPredict on full X.")
                y_pred = model_dict[break_point].predict(X)  # data leakage!!!!
                model_scores[break_point] = metric(y, y_pred)

            elif break_point != x_seq[-1]:
                # predict and evaluate on the next sequence
                next_break_point = x_seq[i + 1]
                if verbose:
                    print(f"\tPredict on {break_point}:{next_break_point}.")
                y_pred = model_dict[break_point].predict(
                    X.iloc[break_point:next_break_point]
                )  # data leakage!!!!
                model_scores[break_point] = metric(
                    y.iloc[break_point:next_break_point], y_pred
                )
            else:
                if verbose:
                    print("\tNo more Data to predict on!")

            model_times[break_point] = time.time() - tic
            model_mem[break_point] = tracemalloc.get_traced_memory()[1] / 1024

        tracemalloc.stop()
        return model_times, model_scores, model_mem, model_dict

    else:
        # River Model: Start of Train and Eval Loop
        if verbose:
            print("Starting Loop for River Model")
        for i, break_point in enumerate(x_seq):

            model_dict[break_point] = model
            if verbose:
                print(f"{i}. Breaking Point: {break_point}")

            gc.collect()
            tracemalloc.reset_peak()
            tic = time.time()

            # -- TRAINING --
            if fit_on_fixed:
                # train on a fixed data set of size 0:fixed_train_size
                if verbose:
                    print(f"\tFit on 0:{fixed_train_size}.")
                model_dict[break_point].learn_many(
                    X.iloc[:fixed_train_size], y.iloc[:fixed_train_size]
                )

            elif fit_on_available_data or i < n_fit:
                # train on the full set of seen data, that is available until train_size, i.e., from 0 to train_size:
                if verbose:
                    print(f"\tFit on 0:{break_point}.")
                model_dict[break_point].learn_many(
                    X.iloc[:break_point], y.iloc[:break_point]
                )

            else:
                # train on the last n_fit partitions only. This is a moving window of size (n_fit x partition size).
                if verbose:
                    print(f"\tFit on {x_seq[i - n_fit]}:{break_point}.")
                model_dict[break_point].learn_many(
                    X.iloc[x_seq[i - n_fit] : break_point],
                    y.iloc[x_seq[i - n_fit] : break_point],
                )

            # -- EVALUATION --
            if eval_on_full_data:
                # predict and evaluate on the full data set:
                if verbose:
                    print("\tPredict on full X.")
                y_pred = model_dict[break_point].predict_many(X)  # data leakage!!!!
                model_scores[break_point] = metric(y, y_pred)

            elif break_point != x_seq[-1]:
                # predict and evaluate on the next sequence
                next_break_point = x_seq[i + 1]
                if verbose:
                    print(f"\tPredict on {break_point}:{next_break_point}.")
                y_pred = model_dict[break_point].predict_many(
                    X.iloc[break_point:next_break_point]
                )  # data leakage!!!!
                model_scores[break_point] = metric(
                    y.iloc[break_point:next_break_point], y_pred
                )
            else:
                if verbose:
                    print("\tNo more Data to predict on!")

            model_times[break_point] = time.time() - tic
            model_mem[break_point] = tracemalloc.get_traced_memory()[1] / 1024

        tracemalloc.stop()
        return model_times, model_scores, model_mem, model_dict


def eval_online_machine_learning(X, y, model, metric, task="clf"):
    """This methods executes an online machine learning task (every single instance is processed separately)

    Args:
        X (DataFrame): _description_
        y (DataFrame): _description_
        model: model
        metric (_type_): metric
        task (str, optional): task_description_. Defaults to "clf".

    Raises:
        Exception: _description_

    Returns:
        dict: model_times
        dict: model_scores
        dict: model_mem
        model: model
    """

    model_times = {}
    model_scores = {}
    model_mem = {}
    scaler = river_preprocessing.StandardScaler()
    i = 0
    y_true = []
    y_pred = []

    tracemalloc.start()
    gc.collect()
    for xi, yi in river_stream.iter_pandas(X, y):
        tic = time.time()

        xi_scaled = scaler.learn_one(xi).transform_one(xi)
        yi_pred = model.predict_one(xi_scaled)

        tracemalloc.reset_peak()

        model.learn_one(xi_scaled, yi)
        model_mem[i] = tracemalloc.get_traced_memory()[1] / 1024
        model_times[i] = time.time() - tic

        if task == "clf":
            y_pred.append(bool(y_pred))
            y_true.append(bool(yi))
        elif task == "reg":
            y_pred.append(yi_pred)
            y_true.append(yi)
        else:
            raise Exception(f"{task} is an invalid input for task")

        model_scores[i] = metric(y_true, y_pred)
        i = i + 1

    tracemalloc.stop()
    return model_times, model_scores, model_mem, model
